(*---------------------------------------------------------------------------*
  $Change$
  Copyright (c) 2002-2010, James H. Woodyatt
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  
    Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    
    Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
  COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
  OF THE POSSIBILITY OF SUCH DAMAGE. 
 *---------------------------------------------------------------------------*)

(** I/O gadget monad.  *)

(** {6 Overview}

    A specialization of [Cf_gadget] in the {b Cf} library, specifically
    targeted to processing a stream of I/O events in concurrent,
    single-threaded network services.  The interface is the same, except the
    input and output type parameters are removed, and there are no explicit
    [read] and [write] functions.
    
    Evaluating an I/O gadget monad results in the execution of an event loop
    (using the [Nx_poll] module).  The underlying [Cf_flow] object conceptually
    reads I/O events from the system and writes commands to wait for more.
    
    An I/O gadget monad is a continuation monad initialized with an abstract
    unit of work for the scheduler.  Monadic functions are defined for the
    following basic operations:
    - starting a new I/O gadget;
    - creating a new wire (with or without endpoints);
    - sending an event on a wire;
    - guarding a gadget on the reception of events from a sequence of wires;
    - scheduling work to be performed when an I/O event occurs.

*)

(** {6 Types} *)

(** An functionally compositional unit of scheduling work in a gadget. *)
type work

(** A gating sequence produced by the [guard] monad for receiving events in
    a gadget using the [guard] function.
*)
type gate

(** A wire capable of carrying events. *)
type 'x wire

(** A continuation monad for composing logical sequences of scheduling work
    within a specific gadget.  The scheduler may interleave units of work from
    different gadgets in any order.
*)
type 'a t = (work, 'a) Cf_cmonad.t

(** A continuation monad for composing a [gate], i.e. sequences of wires to
    block a gadget when waiting for an event.
*)
type 'a guard = (gate, 'a) Cf_cmonad.t

(** {6 Functions and Classes} *)

(** Use [start m] to start a new gadget [m]. *)
val start: unit t -> unit t

(** Use [guard g] to evaluate the guard continuation monad [g] to block the
    current gadget until an event is received on a wires in the sequence used
    in the composition of the continuation.  If none of the wires in the guard
    are still connected to other gadgets, then the scheduler terminates all
    further processing of the gadget.
    
    Control will never pass to any continuation bound to the result.  Instead,
    control will pass to the continuation of the first gating wire in the guard
    to receive a message.
*)
val guard: unit guard -> 'a t

(** Use [abort] to abort the current work unit and return to the scheduler.
    Control will not be passed to any continuation bound to the result.
*)
val abort: 'a t

(** Use [wire] to return a new wire from the scheduler. *)
val wire: 'a wire t

(** Use [wirepair] to return a pair of new wires from the scheduler. *)
val wirepair: ('a wire * 'b wire) t

(** Use [null] to construct a wire that discards every message transmitted
    without ever delivering it.  Such wires can be useful for default arguments
    to some gadget functions.
*)
val null: 'a wire t

(** The class type of wire connectors.  Both receivers and transmitters are
    subtypes of this class type.
*)
class type connector =
    object
        (** For a connector [c], [c#check] returns [false] if the other end of
            the wire has been collected by the garbage collector.  If [c#check]
            returns [true], then the other end of the wire may still be
            connected to a gadget.
        *)
        method check: bool
        
        (** This methods returns a string that may be used with physical
            equality comparisons, i.e. [(==)] and [(!=)] to distinguish the
            underlying wire from other wires in the same scheduler.
        *)
        method id: string
    end

(** The class of connectors used for receiving events.  Use [new rx w] to
    construct a receive connector for the wire [w].
*)
class ['a] rx:
    'a wire ->
    object
        inherit connector
        
        (** Produce a guard to receive an event on the underlying wire and
            dispatch it to the given function.  Use the continuation monad
            binding operator [( >>= )] to produce a guard for a sequence of
            wires.
        *)
        method get: ('a -> unit t) -> unit guard
    end

(** The class of connectors used for sending events.  Use [new tx w] to
    construct a transmit connector for the wire [w].
*)
class ['a] tx:
    'a wire ->
    object
        inherit connector
        
        (** Send the given event on the underlying wire. *)
        method put: 'a -> unit t
    end

(** Use [connect m] to obtain the wire from [m] and use it to return a receive
    and transmit connector pair.
*)
val connect: 'a wire t -> ('a rx * 'a tx) t

(** Use [simplex] to obtain a new wire from the scheduler and use it to return
    a receive and transmit connector pair.  This is a convenient abbreviation
    of [connect wire].
*)
val simplex: ('a rx * 'a tx) t

(** A pair of convenience types for representing each end of a bundle of two
    wires used for duplex communication.  By convention, a [pad] comprises a
    receiver for control events and a transmitter for notification events, and
    a [fix] comprises the transmitter for control events and the receiver for
    notification events
*)
type ('x, 'y) pad = 'x rx * 'y tx
type ('x, 'y) fix = 'y rx * 'x tx

(** Use [connectpair m] to get the pair of wires from [m] and use them to
    construct a matching [fix] and [pad] for each of them.
*)
val connectpair: ('x wire * 'y wire) t -> (('x, 'y) fix * ('x, 'y) pad) t

(** Use [duplex] to get a pair of wires from the scheduler and use them to
    construct a matching [fix] and [pad] for each of them.  This is a
    convenient abbreviation of [connectpair wirepair].
*)
val duplex: (('x, 'y) fix * ('x, 'y) pad) t

(** Use [wrap a b w] to start a new gadget that receives events from the
    receive connector [a], sends event to the transmit connector [b], and
    processes the stream of input and output events with the flow [w].
*)
val wrap: 'i #rx -> 'o #tx -> ('i, 'o) Cf_flow.t -> unit t

(** Use [inherit next] to derive a class that implements an intermediate state
    in a machine.
*)
class virtual next:
    object('self)        
        (** The guard evaluated by this state of the machine. *)
        method virtual private guard: unit guard
        
        (** Use [obj#next] to transition the state of the machine by applying
            {!Cf_gadget.guard} [self#guard].
        *)
        method next: 'a. 'a t
    end

(** Use [inherit start] to derive a class to represent the initial state of a
    machine.  It's [start] method initiates the machine with the virtual
    private [guard] method.
*)
class virtual start:
    object('self)
        (** The first guard evaluationed by the machine after starting. *)
        method virtual private guard: unit guard
        
        (** Starts a new gadget. Defined as [start (guard self#guard)]. *)
        method start: unit t
    end

(** Use [create f] to create a duplex channel, and apply [f] to the resulting
    [pad] to obtain the initial state of a machine.  The machine is started
    and the corresponding [fix] is returned.
*)
val create: (('c, 'n) pad -> #start) -> ('c, 'n) fix t

(** Use [createM f] to create a duplex channel, and apply [f] to the resulting
    [pad] to obtain a continuation monad that evaluates to the initial state of
    a machine.  The machine is started and the corresponding [fix] is returned.
*)
val createM: (('c, 'n) pad -> #start t) -> ('c, 'n) fix t

(** The abstract type of the system I/O event multiplexer. *)
type kernel

(** Use [poll e k] to load the I/O event into the polling mux and event queue
    comprising the kernel [k].
*)
val poll: unit t #Nx_poll.event -> kernel -> unit

(** Use [run f s] to construct a new I/O event kernel, apply it to the function
    [f] to obtain a top-level process for the event loop, and start the process
    running with the initial state [s].
*)
val run: (kernel -> unit t) -> unit

(*--- $File$ ---*)
